def interface____differentiated__9() -> None:
    return None

mod = "djangoexample.only_abstract.models"
summary = "__virtual__.mod_4035906997::djangoexample.only_abstract.models::installed_apps=__installed_apps_hash__::significant=1262772787"

import djangoexample.only_abstract.models
