def interface____differentiated__2() -> None:
    return None

mod = "django.contrib.auth.models"
summary = "__virtual__.mod_2289830437::django.contrib.auth.models::installed_apps=__installed_apps_hash__::significant=1922836542"

import django.contrib.auth.models
import django.db.models
ConcreteQuerySet__AbstractUser = django.db.models.QuerySet[django.contrib.auth.models.User]
ConcreteQuerySet__Group = django.db.models.QuerySet[django.contrib.auth.models.Group]
ConcreteQuerySet__Permission = django.db.models.QuerySet[django.contrib.auth.models.Permission]
ConcreteQuerySet__PermissionsMixin = django.db.models.QuerySet[django.contrib.auth.models.User]
ConcreteQuerySet__User = django.db.models.QuerySet[django.contrib.auth.models.User]
Concrete__AbstractUser = django.contrib.auth.models.User
Concrete__Group = django.contrib.auth.models.Group
Concrete__Permission = django.contrib.auth.models.Permission
Concrete__PermissionsMixin = django.contrib.auth.models.User
Concrete__User = django.contrib.auth.models.User
