def interface____differentiated__8() -> None:
    return None

mod = "djangoexample.relations2.models"
summary = "__virtual__.mod_3328248899::djangoexample.relations2.models::installed_apps=__installed_apps_hash__::significant=1229019932"

import django.db.models
import djangoexample.relations2.models
ConcreteQuerySet__Thing = django.db.models.QuerySet[djangoexample.relations2.models.Thing]
Concrete__Thing = djangoexample.relations2.models.Thing
