def interface____differentiated__11() -> None:
    return None

mod = "django.contrib.auth.base_user"
summary = "__virtual__.mod_2833058650::django.contrib.auth.base_user::installed_apps=__installed_apps_hash__::significant=3626250221"

import django.contrib.auth.base_user
import django.contrib.auth.models
import django.db.models
ConcreteQuerySet__AbstractBaseUser = django.db.models.QuerySet[django.contrib.auth.models.User]
Concrete__AbstractBaseUser = django.contrib.auth.models.User
