def interface____differentiated__1() -> None:
    return None

mod = "django.contrib.admin.models"
summary = "__virtual__.mod_2456226428::django.contrib.admin.models::installed_apps=__installed_apps_hash__::significant=3233719984"

import django.contrib.admin.models
import django.db.models
ConcreteQuerySet__LogEntry = django.db.models.QuerySet[django.contrib.admin.models.LogEntry]
Concrete__LogEntry = django.contrib.admin.models.LogEntry
